from typing import Optional

from django.contrib import admin
from django.db.models.query import QuerySet
from django.http import HttpRequest
from jsoneditor.forms import JSONEditor

from df_remote_config.models import ConfigAttribute, ConfigPart
from df_remote_config.schema import DEFAULT_SCHEMA


@admin.register(ConfigAttribute)
class ConfigAttributeAdmin(admin.ModelAdmin):
    list_display = ("name", "value")


@admin.register(ConfigPart)
class ConfigPartAdmin(admin.ModelAdmin):
    list_display = ("name", "attributes_", "sequence")

    def get_queryset(self, request: HttpRequest) -> QuerySet[ConfigPart]:
        return super().get_queryset(request).prefetch_related("attributes")

    def attributes_(self, obj: ConfigPart) -> str:
        return "&".join([str(attr) for attr in obj.attributes.all()])

    def get_readonly_fields(
        self, request: HttpRequest, obj: Optional[ConfigPart] = None
    ) -> list[str]:
        if obj:
            return ["name"]
        else:
            return []

    def get_form(self, request, obj=None, **kwargs):  # type: ignore
        form = super().get_form(request, obj, **kwargs)
        if not obj:
            form.base_fields.pop("json", None)
        else:
            form.base_fields["json"].widget = JSONEditor(
                jsonschema=obj.get_schema() or DEFAULT_SCHEMA
            )
        return form



# Generated by Django 4.2.5 on 2023-10-02 11:02

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("df_remote_config", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="ConfigAttribute",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("value", models.CharField(max_length=255)),
            ],
            options={
                "unique_together": {("name", "value")},
            },
        ),
        migrations.AlterModelOptions(
            name="configpart",
            options={"ordering": ("sequence",)},
        ),
        migrations.RemoveField(
            model_name="configpart",
            name="tags",
        ),
        migrations.DeleteModel(
            name="ConfigTag",
        ),
        migrations.AddField(
            model_name="configpart",
            name="attributes",
            field=models.ManyToManyField(
                blank=True, to="df_remote_config.configattribute"
            ),
        ),
    ]



# Generated by Django 4.2.5 on 2023-10-02 14:11

from django.db import migrations
import django.utils.timezone
import model_utils.fields


class Migration(migrations.Migration):
    dependencies = [
        ("df_remote_config", "0002_configattribute_alter_configpart_options_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="configpart",
            name="created",
            field=model_utils.fields.AutoCreatedField(
                default=django.utils.timezone.now,
                editable=False,
                verbose_name="created",
            ),
        ),
        migrations.AddField(
            model_name="configpart",
            name="modified",
            field=model_utils.fields.AutoLastModifiedField(
                default=django.utils.timezone.now,
                editable=False,
                verbose_name="modified",
            ),
        ),
    ]



from typing import Any

from django.http import HttpRequest, HttpResponse
from django.utils.module_loading import import_string
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import OpenApiParameter, extend_schema
from rest_framework import permissions, views
from rest_framework.exceptions import NotFound

from df_remote_config.settings import api_settings


class RemoteConfigView(views.APIView):
    http_method_names = ["get"]
    permission_classes = [permissions.AllowAny]

    @extend_schema(
        parameters=[
            OpenApiParameter(
                name="part",
                description="Name of the configuration part.",
                required=True,
                type=OpenApiTypes.STR,
                location=OpenApiParameter.QUERY,
                enum=api_settings.PARTS.keys(),
            ),
            OpenApiParameter(
                name="attributes",
                description="json-encoded object of attribute key value pairs",
                required=False,
                type=OpenApiTypes.OBJECT,
                location=OpenApiParameter.QUERY,
            ),
        ],
        responses={
            200: OpenApiTypes.OBJECT,
        },
    )
    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        part = request.GET.get("part", api_settings.DEFAULT_PART)
        if part not in api_settings.PARTS:
            raise NotFound()

        part_config = api_settings.PARTS[part]

        handler = import_string(
            part_config.get("HANDLER_CLASS", "df_remote_config.handlers.DefaultHandler")
        )()
        return handler.handle_request(request, part, *args, **kwargs)



PART_SCHEMAS: dict[str, dict] = {
    "legal": {
        "type": "object",
        "properties": {
            "terms_of_service": {"type": "string"},
            "privacy_policy": {"type": "string"},
        },
        "required": ["terms_of_service", "privacy_policy"],
    },
    "auth": {},
    "app_launch": {
        "type": "object",
        "properties": {
            "app_version": {
                "type": "object",
                "properties": {
                    "required": {"type": "string"},
                    "latest": {"type": "string"},
                },
                "required": ["required", "latest"],
            },
        },
        "required": ["app_version"],
    },
}

DEFAULT_SCHEMA = {
    "type": "object",
    "properties": {},
    "required": [],
    "additionalProperties": True,
}



from django.db import models
from django.utils.module_loading import import_string
from model_utils.models import TimeStampedModel

from df_remote_config.fields import NoMigrationsChoicesField
from df_remote_config.schema import DEFAULT_SCHEMA, PART_SCHEMAS
from df_remote_config.settings import api_settings


class ConfigAttribute(models.Model):  # type: ignore
    name = models.CharField(max_length=255)
    value = models.CharField(max_length=255)

    def __str__(self) -> str:
        return f"{self.name}={self.value}"

    class Meta:
        unique_together = ("name", "value")


class ConfigPartQuerySet(models.QuerySet):
    def filter_attributes(self, attributes: dict[str, str]) -> "ConfigPartQuerySet":
        if not attributes:
            return self

        attribute_condition = models.Q()
        for name, value in attributes.items():
            attribute_condition |= models.Q(name=name, value=value)
        return self.filter(
            attributes__in=ConfigAttribute.objects.filter(attribute_condition)
        )


class ConfigPart(TimeStampedModel):
    objects = ConfigPartQuerySet.as_manager()
    name = NoMigrationsChoicesField(
        max_length=255, choices=[(part, part) for part in api_settings.PARTS]
    )
    sequence = models.PositiveIntegerField(default=1000)
    attributes = models.ManyToManyField(ConfigAttribute, blank=True)
    json = models.JSONField(blank=True, default=dict)

    def __str__(self) -> str:
        return f"ConfigPart<{self.id}>: {self.name}"

    def get_schema(self) -> dict:
        # If the schema is not defined in the settings, use the default schema
        if "SCHEMA" not in api_settings.PARTS[self.name]:
            return PART_SCHEMAS.get(self.name, DEFAULT_SCHEMA)

        return import_string(api_settings.PARTS[self.name]["SCHEMA"])

    class Meta:
        ordering = ("sequence",)



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class DfRemoteConfigConfig(AppConfig):
    name = "df_remote_config"
    verbose_name = _("Django DF Remote Config")

    class DFMeta:
        api_path = "remoteconfig/"



from typing import Any

from django.conf import settings
from rest_framework.settings import APISettings

DEFAULTS: dict[str, Any] = {
    "PARTS": {
        "legal": {
            "HANDLER_CLASS": "df_remote_config.handlers.DefaultHandler",
        },
        "default": {},
        "app_launch": {
            "HANDLER_CLASS": "df_remote_config.handlers.AppLaunchHandler",
        },
    },
    "DEFAULT_PART": "default",
}

api_settings = APISettings(getattr(settings, "DF_REMOTE_CONFIG", None), DEFAULTS)



# Generated by Django 4.2.5 on 2023-10-02 05:00

import df_remote_config.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="ConfigTag",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name="ConfigPart",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    df_remote_config.fields.NoMigrationsChoicesField(max_length=255),
                ),
                ("sequence", models.PositiveIntegerField(default=1000)),
                ("json", models.JSONField(blank=True, default=dict)),
                (
                    "tags",
                    models.ManyToManyField(blank=True, to="df_remote_config.configtag"),
                ),
            ],
        ),
    ]



from typing import Optional

from django.db.models import Max
from django.http import HttpRequest
from rest_framework.exceptions import NotFound
from rest_framework.response import Response

from df_remote_config.models import ConfigPart


class AbstractHandler:
    def handle_request(self, request: HttpRequest, part_name: str) -> Response:
        raise NotImplementedError


class DefaultHandler(AbstractHandler):
    def get_config_part(
        self, request: HttpRequest, part_name: str
    ) -> Optional[ConfigPart]:
        attributes = request.GET.dict()
        attributes.pop("part", None)

        return (
            ConfigPart.objects.filter(name=part_name)
            .filter_attributes(attributes)
            .first()
        )

    def get_part_data(self, part: ConfigPart) -> dict:
        return part.json

    def handle_request(self, request: HttpRequest, part_name: str) -> Response:
        if config_part := self.get_config_part(request, part_name):
            return Response(self.get_part_data(config_part))
        else:
            raise NotFound()


class AppLaunchHandler(DefaultHandler):
    def get_part_data(self, part: ConfigPart) -> dict:
        data = super().get_part_data(part)

        data["updates"] = {
            "parts": {
                part["name"]: part["modified"].isoformat()
                for part in ConfigPart.objects.values("name").annotate(
                    modified=Max("modified")
                )
            }
        }

        return data
